from .. import Substitution, Constraint, generate, init_grammar
import exrex
import funcy as fc
import itertools

C, S = Constraint, Substitution

eng, tptp = "eng","tptp"

ADJECTIVES = ['rich', 'quiet', 'old', 'tall', 'kind', 'brave', 'wise',
              'happy', 'strong', 'curious', 'patient', 'funny', 'generous', 'humble']

NAMES = ['mary', 'paul', 'fred', 'alice', 'john', 'susan', 'lucy']

ADJS=['rich','quiet','old','tall','kind','brave','wise','happy','creative',
    'strong','curious','patient','funny','organized','formal','humble']
ADJS+=['left_handed','curly_haired','popular','romantic','blue_eyed','long_haired','scarred','colorblind']
    

def FOL_grammar(N_PREMS=8, names=NAMES, adjs=ADJS, empty_room=True):
    
    R=init_grammar(['tptp','eng'])
    R('start(setup)', '0')
    
    def render_setup(x):
        persons=x.state['persons']
        room=x.state['room']
        persons_str = ', '.join([p.title() for p in persons])
        eng_setup=f"{persons_str} are the only persons in the {room}.\n{x[0]@eng}"
        if len(persons)==1:
            eng_setup=eng_setup.replace(' are ',' is ').replace('persons','person')
        in_room="&".join([f"room({x})" for x in persons])
        disj="|".join([f"X='{x}'" for x in persons])

        dist = ('&'.join('({}!={})'.format(*a) for a in itertools.product(persons,persons) if len(set(a))!=1))
        tptp_setup=in_room + f'&(dist)' + f"&(![X]:(room(X)=>({disj})))&\n{x[0]@tptp}" 
        return tptp_setup,eng_setup
    
    room = 'room'
    
    def pluralize(s):
        s=s.replace('does','do')
        s=s.replace('is a','are')
        s=s.replace('is ','are ')
        s=s.replace('es ','e ')
        s=s.replace('person','persons')
        return s

        
    psums = [list(names[:i]) for i in range(1, int(len(names)*0.8))]
    
    for persons in psums:
        R('setup(valid_block)', lambda x:render_setup(x)[0], lambda x:render_setup(x)[1],
            vars=dict(persons=persons,room=room))
    
    def rs(x):
        return f"(there_is_a_room)&\n{x[0]@tptp}", f"there is a room.\n{x[0]@eng}"

    if empty_room:
        for persons in [names[:4],[],[]]:
            R('setup(valid_block)', lambda x:rs(x)[0], lambda x:rs(x)[1],
                vars=dict(persons=[],room=room), weight=len(psums))
    
    def neg_constraint(x):
        """avoids material negation issues"""
        for d in x.descendants:
            if type(d.rule)==str:
                continue
            tptp=str(d.rule.templates['tptp'])
            if "=>" in tptp or "<=" in tptp:
                return False
            eng=str(d.rule.templates['eng'])
            if 'who' in eng:
                return False
        return True

    def no_free_var(x):
        s=x[0]@tptp
        return 'X' not in s or set('?!')&set(s)
    R('valid_block(block)','0', constraint=no_free_var) #🏁
    
    nesting_limit = lambda x: x[0].render('eng').count('it is ') <= 2
    R("term(term)", "~(0)", "it is not the case that “0”", constraint=nesting_limit,
      state_constraint=neg_constraint,weight=1)
    
    R("term(term)", "0", "it is true that “0”",constraint=nesting_limit,weight=0.25)
    
    
    R("term(term)", "(mary_dream=>(0))", "Mary dreamt that “0”",constraint=nesting_limit,weight=0.15)    

    
    #R("block(term)", "0", "0",weight=1)
    
    def block_cond(x):
        for d in x.descendants:
            if type(d.rule)!=str and d.rule.name in ['cproperty','term']:
                if '=' in str(d.rule.templates['tptp']):
                    return False
                if 'who' in str(d.rule.templates['eng']):
                    return False
        return True
    
    R('nterm(term)','0', state_constraint=block_cond)
    
    if True:
        
        R("cond(nterm,term)", "(0)=>(1)", "if “0” then “1”", weight=4)
    
        #R("cond(nterm,nterm,term)", "((0)=>(1))&\n((1)=>(2))", "if “0” then “1”\nif “1” then “2”")
        R("cond(nterm,term)", "(0)=>(1)", "“1” if “0”")
        R("cond(nterm,nterm)", "(0)<=>(1)", "“1” if “0” and vice versa")
        R("cond(nterm,term)", "(0)=>(1)", "“0” only if “1”")
        R("cond(nterm,nterm)", "(0)<~>(1)", "either “0” or “1” but not both")
        R("cond(term,nterm)", "~(1)=>(0)", "“0” unless “1”")
        R('cond(nterm,term,term)', '((0)=>(1))&((~(0))=>(2))', 'if “0” then “1” otherwise “2”')
        #R("block(term,term)", "(0)&\n(1)", "0\n1")
    
        conj_tptp = lambda i: '&\n'.join([f'({i})' for i in range(i)])
        conj_eng  = lambda i: '\n'.join([f'{i}' for i in range(i)])
        conj = lambda i: (conj_tptp(i), conj_eng(i))
    
        
        for n in [N_PREMS]:
            for nconds in range(min(10,int(n//2))): #1.5 
                nterms = n - nconds  
                arg1 = "block(" + "cond,"*nconds + ("term,"*(nterms-1)).rstrip(',') + ")"
                w=(2 if nconds==0 else 1)
                w=w*(0.15 if n<=8 else 1)
                R(arg1, *conj(n-1),weight=w)

        
    props = ['propositiona', 'propositionb', 'propositionc','propositiond','propositione']
    props=[]
    for p in props:
        R('prop',p)
        R('prop',"~"+p)
    
    R('term(prop)','0')
    R('term(prop,prop)','((0)&(1))','“0” and “1”')
    R('term(prop,prop,prop)','((0)&(1)&(2))','“0” and “1” and “2”')
    R('term(prop,prop)','((0)|(1))','“0” or “1” or both')
    R('term(prop,prop)','((0)<~>(1))','“0” or “1” but not both')
    R('term(prop,prop)','~((0)|(1))','neither “0” nor “1”')
    
    for n in names:
        R("entity_",n,n.title())
    
    def inside_person(x):
        persons =x[0].state.get('persons','')
        outcome= x[0]@tptp in persons
        return outcome
    
    entity_constraint = None
    R('entity(entity_)','0','0')#,constraint=entity_constraint)
    

    for adj in adjs:
        if adj=="rich":
            continue
        R('adjective', adj,weight=0.8)
    
    R('noun','person')
    R('adj_noun(adjective_chain,noun)','0&1(?)','0 1')
    R('adjective_chain(adjective_chain,adjective)','0&1(?)','0 1',constraint=C('1∉0'))
    R('adjective_chain(adjective)','0(?)','0')
    
    #R('outside_entity(entity_)','0','0',constraint=lambda x: not inside_person(x))
    
    R('property(adj_noun)','0','is a 0',weight=2)
    R('property(adj_noun)','~(0)','is not a 0',weight=1)
    
    R('neg_adj(neg_adj,adjective)','0&~1(?)','0, not 1',constraint=C('1∉0'))
    R('neg_adj(adjective)','~0(?)','not 0')
    R('property(neg_adj)','0','is 0',weight=0.5)
    
    
    R('mneg(mneg)','~0', 'not 0',weight=0.75)
    R('mneg(adjective)','0(?)','0')
    R('property(mneg)','0','is 0',weight=0.5)
    
    
    R('property(adjective)','0(?)','is 0')
    
    R('property(sentiment,outside_entity)', '~0(1,?)','does not 0 1',weight=0.5)
    R('property(sentiment,outside_entity)', '~0(?,1)','is not 0d by 1',weight=0.5)
    # /negations
    
    R('term(entity,entity,adjective,adjective)','(2(0))&(3(1))','0 and 1 are respectively 2 and 3')  
    

    R("sentiment", "like"), R("sentiment", "hate")
    R('property(sentiment,outside_entity)', '0(1,?)','0s 1',weight=0.25)
    R('property(sentiment,outside_entity)', '0(?,1)','is 0d by 1',weight=0.25)
    R('term(entity,outside_entity,sentiment)','(2(0,1) & 2(1,0))', '0 and 1 2 each other')
    R('term(outside_entity,entity,sentiment)','(2(0,1) & 2(1,0))', '0 and 1 2 each other')
    
    R('relation(sentiment)','0'), R('relation(predicate)','0')
    R('term(property,property,relation)',r'\?[X,Y]:((0[?←X])&(1[?←Y])&(2(X,Y)))','someone who 0 2 someone who 1')
    # other quantifiers are ambiguous
    
    preds = list(exrex.generate('pred[a-j]'))

    for i in preds:
        R('atomic_predicate',i)
    R('property(atomic_predicate)', '0(?)','0',weight=6)
    R('property(atomic_predicate)', '~0(?)','~0',weight=1)
    
    R('neg_property(atomic_predicate)', '~0(?)','~0')
    
    R('term(entity,property)', '1[?←0]',
      fc.rcompose(S('0 1', lang='eng'), lambda x: x.replace('they', 'he/she')),
    constraint=[C('0∉1,1∉0'), lambda x: 'who' not in x[1]@eng],
      weight=8)

    R('term(entity,property,property)','(1[?←0])&(2[?←0])','0 who 1 2',weight=1)
    
    R('term(property,entity,entity)', '0[?←1]&0[?←2]', fc.rcompose(S('1 and 2 0',lang='eng'),pluralize),
        constraint=[
            C('0∉1,1∉2,0∉2'),
            lambda x:'pred' not in x[0]@tptp,
            lambda x:'who' not in x[0]@eng],
      weight=2)
    
    R('quantifier', '!', 'everyone',weight=10), 
    R('quantifier','~!','not everyone', state_constraint=neg_constraint)
    ####################R('quantifier','~\?','no one', state_constraint=neg_constraint) 
    R('group','room','in the room',weight=8), R('group','~room','outside the room')
    R('group','anywhere','anywhere')
    
    R('X_quantifier(quantifier,group)','0[X]:(1(X)=>(?))','0 1')
    
    
    R('term(adjective,adjective,group)', '![X]:(2(X)=>(0(X)=>1(X)))','all 0 persons 2 are 1',weight=2)
    R('term(adjective,adjective,group)', '![X]:(2(X)=>(0(X)=>~1(X)))','no 0 person 2 is 1')
    
    def block_property(x):
        for d in x.descendants:
            if d.rule and 'property' in d.rule.args:
                return False
        return True
    
    kw=dict(weight=0.3,state_constraint=block_property)
    R('property(property,property)', '((0)<~>(1))', 'either 0 or 1 but not both',**kw)
    R('property(property,property)', '((0)|(1))', '0 or 1 or both',**kw)
    R('property(property,property)', '((0)&(1))', '0 and 1',**kw)
    R('property(property,property)', '~((0)|(1))', 'neither 0 nor 1',**kw)
    R('property(property,property,property)', '((0)&(1)&(2))', '0, 1 and 2',**kw)
    R('property(property,property,property)', '((0)|(1)|(2))', '0, 1 or 2',**kw)
    
    R('cproperty(property,property)', '((0)=>(1))', 'who 0 1',weight=4)
    R('cproperty(property,property)', '((0)=>(1))', '1 if they 0',weight=4)
    R('cproperty(property,property)', '((1)<=(0))', '0 only if they 1')
    R('cproperty(property,property)', '((0)<=>(1))', '0 if they 1 and vice versa',weight=0.5)
    R('cproperty(property,property)', '((0)<=>(1))', '0 if and only if they 1',weight=0.5)
    #unless, otherwise
    
    #implication as disjunction for hypothesis (more explicit)
    R('cproperty(neg_property,property)', '((0)|(1))', '1 or 0',weight=0.5,state_constraint=block_property) 
    R('cproperty(property,neg_property)', '((0)|(1))', '1 or 0',weight=0.5,state_constraint=block_property) 
    
    
    R('X_property(cproperty)','0[?←X]','0')
    R('X_property(property)','0[?←X]','0',weight=0.1)
    
    R('term(X_quantifier,X_property)', '0[?←1]', '0 1', weight=12) 

    R('E_quantifier(group)',r'\?[X]:(0(X)&(?))','someone 0')
    R('E_property(property)','0[?←X]','0')
    R('term(E_quantifier,E_property)', '0[?←1]', '0 1', weight=1) 
    
    R('xp(property)','0[?←X]','0')
    R('term(xp,xp)', '(![X]:((0)=>(1)))', 'if someone 0 then he/she 1')
    R('term(xp,xp)', '(![X]:((0)<=>(1)))', 'if someone 0 then he/she 1 and vice versa')
    
    ws=0.5
    R('term(property,group)',r'((\?[X]:(1(X)&0[?←X])))','at least one person 1 0',weight=ws)
    R('term(property,group)',r'((\?[X]:(1(X)&0[?←X]))&(![X,Y]:((1(X)&1(Y)&(0[?←X])&(0[?←Y]))=>(X=Y))))','only one person 1 0',weight=ws)
    R('term(property,group)',r'(\?[X,Y]:(1(X)&1(Y)&(0[?←X]&0[?←Y])&(X!=Y)))','more than one person 1 0',weight=ws)
    

    def at_least(n,cond, prop):
        assert n<25 # relies on letters variables
        variables = [chr(ord('A')+i) for i in range(n)]
        conjuncts = [f'{prop}[?←{var}]' for var in variables]
        inequalities = [f'({variables[i]}!={variables[j]})' for i in range(n) for j in range(i+1, n)]
        conds = [f'{cond}({var})' for var in variables]
        formula = '&'.join(conds+inequalities+conjuncts)
        return f'?[{",".join(variables)}]:({formula})'
    
    def at_most(n,cond, prop):
        return f"~({at_least(n+1,cond,prop)})"
    
    def exactly(n,cond, prop):
        return f"({at_least(n,cond,prop)})&({at_most(n,cond,prop)})"
    
    ws = 0.5/10
    for exp in ['at_least','at_most','exactly']:
        for i in range(2,5+1):
            n={1:'one',2:'two',3:'three',4:'four',5:'five'}[i]
            R('term(property,group)',eval(exp)(i,1,0), f'{exp.replace("_"," ")} {n} persons 1 0',weight=ws)


    def hyp_cst(x):
        for d in x.descendants:
            if type(d.rule)==str:
                continue
            tptp=str(d.rule.templates['tptp'])
            if "=>" in tptp or "<=" in tptp:
                return False
            if 'room' in tptp:
                return False
        return True
    
    R('hypothesis(term)', '0', '0',  state_constraint=[neg_constraint, hyp_cst],
      vars=dict(persons=persons,room=room))


    
    return R


from .generate_recursive import generate_recursive
from .generate_sequential import generate_sequential
import random

def generate(start, depth=10, n_iter=10_000, mode='sequential', seed=None, min_depth=None, *args, **kwargs):
    random.seed(seed)
    if type(start)==type:
        start=start.start()

    modes = {'recursive': generate_recursive,
              'sequential': generate_sequential,
            }

    generate = modes[mode]    
    """Generate one production using specified mode."""
    for _ in range(n_iter):
        result = generate(start, max_depth=depth, min_depth=min_depth, *args, **kwargs)
        if min_depth and result and result[0].height < min_depth:
            continue
        if result: return result[0]
    raise ValueError('Incomplete generation')


# generate.py
import random
from collections import deque, ChainMap
from functools import lru_cache
import anytree
from easydict import EasyDict as edict
# Assuming .grammar and other imports are correctly set up in the project structure
from .grammar import Substitution

import re

# ... (FastProduction class and _precompute_height_bounds function remain unchanged) ...
class FastProduction:
    """A unified, high-performance production node for all generation modes."""
    __slots__ = ('rule', 'type', 'parent', 'children', 'local_state', 'cache', '_depth', 'step', 'save', '_height')

    def __init__(self, rule=None, type=None, parent=None, children=None, state=None):
        self.rule, self.type, self.parent, self.cache = rule or '', type or (rule.name if rule else ''), parent, {}
        self.children = children if children is not None else []
        self._depth = (parent.depth + 1 if parent else 0)
        self._height = -1
        self.local_state = state or {}

        if self.rule and not children:
            if self.rule.state:
                self.local_state = {**self.rule.state, **self.local_state}
            self.children = [FastProduction(type=arg, parent=self) for arg in self.rule.args]
            
    @property
    def state(self):
        if self.parent:
            return ChainMap(self.local_state, self.parent.state)
        return self.local_state

    def clone(self):
        new_root = FastProduction(rule=self.rule, state=self.local_state.copy())
        if hasattr(self, 'step'): new_root.step = self.step
        if hasattr(self, 'save'): new_root.save = self.save
        q_orig, q_clone = deque([self]), deque([new_root])
        while q_orig:
            orig_node, clone_node = q_orig.popleft(), q_clone.popleft()
            clone_node.children = [FastProduction(rule=c.rule, type=c.type, parent=clone_node, state=c.local_state.copy()) for c in orig_node.children]
            q_orig.extend(orig_node.children); q_clone.extend(clone_node.children)
        return new_root
    
    def _find_first_unexpanded_leaf(self):
        q = deque([self])
        while q:
            node = q.popleft()
            if not node.rule:
                return node
            q.extendleft(reversed(node.children))
        return None

    def update_height(self):
        if self._height != -1: return self._height
        self._height = 0 if not self.children else 1 + max((c.update_height() for c in self.children), default=-1)
        return self._height

    @property
    def height(self):
        if self._height == -1: self.update_height()
        return self._height
        
    @property
    def depth(self): return self._depth
    @property
    def leaves(self):
        if not self.children: return [self]
        leaves_list, q = [], deque([self])
        while q:
            node = q.popleft()
            if not node.children: leaves_list.append(node)
            else: q.extendleft(reversed(node.children))
        return leaves_list
    @property
    def ancestors(self):
        anc, node = [], self.parent
        while node: anc.append(node); node = node.parent
        return tuple(anc)
    @property
    def siblings(self):
        return tuple(c for c in self.parent.children if c is not self) if self.parent else tuple()
    @property
    def descendants(self):
        if not self.children: return tuple()
        desc, q = [], deque(self.children)
        while q:
            node = q.popleft(); desc.append(node); q.extend(node.children)
        return tuple(desc)

    def check(self, mode='args'):
        if not self.rule: return True
        if mode == 'args':
            if not all(x.rule for x in self.children): return True
            return all(c(self.children) for c in self.rule.constraint)
        if mode == 'state':
            return all(c(x) for x in [self] + list(self.ancestors) for c in x.rule.state_constraint)
        return True

    def render(self, lang=None):
        if lang in self.cache: return self.cache[lang]
        if not self.rule: return "#" + self.type
        template = self.rule.templates.get(lang)
        if template is None: return "#" + self.type
        
        if callable(template):
            args, template_func = self.children, template
        else: # String template
            if '?←' in template:
                args = self.children
                template_func = Substitution(template, lang)
            else:
                args = [c.render(lang) for c in self.children]
                template_func = template.format

        out = template_func(*args)
        if "#" not in out: self.cache[lang] = out
        return out

    def __getitem__(self, key): return self.children[key]
    __matmul__ = render

    def __repr__(self):
        if self.rule: return f"PROD:{self.type}({', '.join(self.rule.args)})"
        else: return f"PROD:{self.type}"
    
    def dict(self, use_cls=False):
        res = edict({l: self@l for l in self.rule.langs})
        if use_cls: res.cls = self
        return res

    def to_anytree(self):
        memo = {}
        def _convert_node(node):
            if id(node) in memo: return memo[id(node)]
            new_node = anytree.Node(node.type)
            new_node.rule = node.rule
            new_node.state = dict(node.state) 
            for child in node.children:
                _convert_node(child).parent = new_node
            memo[id(node)] = new_node
            return new_node
        return _convert_node(self)

@lru_cache(maxsize=None)
def _precompute_height_bounds(Rule):
    non_terminals = list({r.name for r in Rule._instances})
    min_h, max_h = {nt: 999 for nt in non_terminals}, {nt: 0 for nt in non_terminals}
    for r in Rule._instances:
        if not r.args: min_h[r.name], max_h[r.name] = 0, 0
    for _ in range(len(non_terminals) + 1):
        for r in Rule._instances:
            if not r.args: continue
            child_mins = [min_h.get(arg, 999) for arg in r.args]
            child_maxs = [max_h.get(arg, 0) for arg in r.args]
            if all(h != 999 for h in child_mins): min_h[r.name] = min(min_h[r.name], 1 + max(child_mins))
            if all(h != 0 or max_h.get(arg, -1) != -1 for arg, h in zip(r.args, child_maxs)):
                max_h[r.name] = max(max_h[r.name], 1 + max(child_maxs))
    min_h = {k: v for k, v in min_h.items() if v != 999}
    return min_h, max_h


def generate_sequential(start, k=1, max_depth=12, min_depth=None, bushiness=1.0,
                        skip_check=False, max_steps=5000, save_prob=0.0125, debug=False, **kwargs):
    min_depth = min_depth or 0
    if min_depth > max_depth: return []

    Rule = type(start)
    min_heights, max_heights = _precompute_height_bounds(Rule)

    def save(p, s, st):
        ckpt = p.clone(); ckpt.step, ckpt.save = st, 1
        s.insert(([0] + [i for i, x in enumerate(s) if hasattr(x, 'save') and x.save])[-1], ckpt)

    def s_choices(seq, w, k):
        if not seq: return []
        if w and sum(w) > 0: return random.choices(seq, weights=w, k=k)
        return random.sample(seq, min(k, len(seq)))

    start_prod = FastProduction(start)
    start_prod.local_state['target_min_depth'] = min_depth
    start_prod.step, start_prod.save = 0, 0
    
    stack, step = [start_prod], 0
    while stack:
        step += 1
        if step > max_steps: return []
        prod = stack.pop()

        lv = prod._find_first_unexpanded_leaf()

        if not lv:
            prod.update_height()
            if min_depth <= prod.height <= max_depth and \
               (skip_check or all(x.check() for x in [prod] + list(prod.descendants))):
                return [prod]
            continue

        current_target_min_depth = lv.state.get('target_min_depth', 0)
        if debug: print(f"[{step}] Expanding '{lv.type}' @ d={lv.depth} (goal_min={current_target_min_depth}) | Stack: {len(stack)}")
            
        # --- RULE SELECTION AND PRUNING ---
        potential_rules = Rule.get_rules(lv.type, shuffle=True)
        valid_non_terminals, valid_terminals = [], []

        for r in potential_rules:
            # Min height from this rule to a leaf
            min_h_rule = 0 if not r.args else 1 + max((min_heights.get(a, max_depth + 1) for a in r.args), default=0)
            # Max height this rule could possibly produce
            max_h_rule = 0 if not r.args else 1 + max((max_heights.get(a, 0) for a in r.args), default=0)

            # Hard Pruning:
            # 1. Don't exceed max_depth
            if lv.depth + min_h_rule > max_depth: continue
            # 2. Must be able to reach the target depth for this branch
            if lv.depth + max_h_rule < current_target_min_depth: continue
            
            if r.args:
                valid_non_terminals.append(r)
            else: # Terminal rule
                # 3. If we pick a terminal, this leaf's depth must meet the target
                if lv.depth >= current_target_min_depth:
                    valid_terminals.append(r)
        
        # --- DECIDE WHETHER TO TERMINATE OR CONTINUE ---
        rules_to_consider = []
        # If the branch has met its depth goal and we can terminate, we should
        # prefer to do so, especially with low bushiness.
        can_terminate = len(valid_terminals) > 0
        has_met_goal = lv.depth >= current_target_min_depth

        # With probability (1-bushiness), force termination if possible.
        # This creates "vines" for low bushiness.
        if has_met_goal and can_terminate and random.random() > bushiness:
             rules_to_consider = valid_terminals
        else:
             rules_to_consider = valid_non_terminals + valid_terminals

        if debug: print(f"  Candidates ({len(rules_to_consider)}): {[r.name for r in rules_to_consider]}")
        if not rules_to_consider: continue
            
        rules_to_try = s_choices(rules_to_consider, [r.weight for r in rules_to_consider], k)
        if len(rules_to_try) > 1 and random.random() < save_prob: save(prod, stack, step)
        
        for i, rule in enumerate(rules_to_try):
            branch = prod if i == len(rules_to_try) - 1 else prod.clone()
            
            path, curr = [], lv
            while curr.parent: path.append(curr.parent.children.index(curr)); curr = curr.parent
            target_lv = branch
            for idx in reversed(path): target_lv = target_lv.children[idx]

            target_lv.rule = rule
            if rule.state: target_lv.local_state.update(rule.state)

            if rule.args:
                # One child MUST carry the torch for the depth requirement.
                critical_candidates = [
                    idx for idx, arg_type in enumerate(rule.args)
                    if target_lv.depth + 1 + max_heights.get(arg_type, 0) >= current_target_min_depth
                ]
                if not critical_candidates: continue # Should not happen due to prior checks
                
                critical_idx = random.choice(critical_candidates)
                
                new_children = []
                for idx, arg_type in enumerate(rule.args):
                    child = FastProduction(type=arg_type, parent=target_lv)
                    if idx == critical_idx:
                        # This child inherits the same absolute depth target
                        child.local_state['target_min_depth'] = current_target_min_depth
                    else:
                        # Other children get a "bushiness"-controlled target.
                        # For b=0, target is own depth (shallow). For b=1, target is the main target (bushy).
                        remaining_depth = max(0, current_target_min_depth - child.depth)
                        bushy_addon = int(bushiness * remaining_depth)
                        child.local_state['target_min_depth'] = child.depth + random.randint(0, bushy_addon)
                    new_children.append(child)
                target_lv.children = new_children
            else:
                 target_lv.children = []


            if not skip_check and (not target_lv.check('state') or (all(s.rule for s in target_lv.siblings) and not target_lv.parent.check('args'))):
                continue
            
            stack.append(branch)
            break
            
    return []


from .FOL import FOL_grammar
from .tinypy import tinypy_grammar
from ..grammar import init_grammar
import random
import string

def arith():
    g = init_grammar(['py'], name="arithmetics", preprocess_template=lambda s:s)
    g('start(expr)',        '{0}')
    g('expr(expr)',       '({0})',            weight=1)
    g('expr(expr,expr)',  '{0} + {1}', weight=2)
    g('expr(expr,expr)',  '{0} - {1}', weight=1)
    g('expr(expr,expr)',  '{0} * {1}')
    g('expr(expr,expr)',  '{0} / {1}')
    g('expr(expr)',       '({0})**2',         weight=.25)
    g('expr(value)',       '{0}',weight= 10)
    g('value',  'NUM')

    return g


def regex_grammar():

    from faker import Faker
    fake = Faker()
    wordlist = fake.words(nb=100,unique=True)
    
    R = init_grammar(['re'], preprocess_template=lambda x: x)

    R('start(regex)', '{0}')
    R('regex(regex,regex)', '{0}{1}', weight=2)
    R('regex(regex)', '({0})', weight=2)
    R('regex(regex,regex)', '{0}|{1}', weight=2)
    R('regex(char)', '{0}',weight=1)
    R('regex(word)', '{0}',weight=3)

    for w in random.sample(wordlist, 8):
        R('word', w)

    R('regex(regex)?', '{0}?')
    R('regex(regex)*', '{0}*')
    R('regex(regex)+', '{0}+')
    R('regex(rangechar,rangechar)', '[{0}-{1}]')
    R('regex(predef)', '{0}',weight=3)

    chars = string.ascii_letters + string.digits
    for c in chars:
        R('char', c)
        R('rangechar', c)

    for s in [r'\d', r'\w', r'\s', '.', r'\.']:
        R('predef', s, weight=1)

    for s in [r'\D', r'\W', r'\S', r'\\', r'\(', r'\)', r'\[', r'\]', r'\t', r'\n']:
        R('predef', s, weight=0.25)

    return R


def negate_predicate(predicate):
    replacements = {
        "is ": "is not ",
        "has ": "does not have ",
        "does ": "does not ",
        'can ': 'cannot '
    }
    for key, val in replacements.items():
        if predicate.startswith(key):
            return predicate.replace(key, val, 1)
    
    words = predicate.split(" ")
    if words[0].endswith("s"):
        words[0] = words[0][:-1]  # Remove 's' from the verb
    return "does not " + " ".join(words)



short_propositions = [
    "Planet Xylos has diamond rain.",
    "Bellbridge's houses are all purple.",
    "Gravity inverts in Oakhaven on Tuesdays.",
    "A tree in Whispering Woods has golden fruit.",
    "A square cloud is over Silver Lake.",
    "Blackwood Manor's mirrors show the future.",
    "Avani's people eat only blue moss.",
    "A singing flower blooms in the Amazon.",
    "John Smith's car runs on ethanol.",
    "Eldoria's children talk to animals."
]




neg_short_propositions = [
    "Planet Xylos has no diamond rain.",
    "Not all Bellbridge's houses are purple.",
    "Gravity does not invert in Oakhaven on Tuesdays.",
    "No tree in Whispering Woods has golden fruit.",
    "No square cloud is over Silver Lake.",
    "Blackwood Manor's mirrors do not show the future.",
    "Avani's people do not eat only blue moss.",
    "No singing flower blooms in the Amazon.",
    "John Smith's car does not run on ethanol.",
    "Eldoria's children do not talk to animals."
]
 
#are any of those logically entailing or contradicting each other in a clear-cut way?
# I'm not talking about slight contrast, correlation or negataive correlation

predicates = [
    "is a client of Meta", "is a client of Costco", "is a client of LVMH", "uses an ios phone",
    "owns an Android phone", "owns a smart tv", "watches fantasy movies",
    "reads mystery novels", "writes a travel blog", "practices digital art",
    "enjoys virtual reality gaming", "is a drone photographer", "enjoys making ceramics", "enjoys watercolor painting", "practices graffiti art",
    "writes poetry", "practices calligraphy", "enjoys landscape photography", "enjoys macrame", "enjoys origami",
    "can play the piano", "can play the guitar",
    "can play the harmonica", "can play the flute", "plays the drums", "plays the violin", "enjoys salsa dancing", "does enjoy trail running",
    "does enjoy mountain biking", "enjoys snowboarding", "enjoys spelunking", "enjoys cross-country skiing",
    "enjoys stand-up paddleboarding", "enjoys windsurfing", "practices pilates",
    "practices tai chi", "practices zumba", "practices archery", "practices kickboxing",
    "enjoys skydiving", "collects foreign coins", "collects vintage stamps", "collects modern art",
    "collects vintage vinyl records", "collects action figures", "collects antique clocks",
    "collect rare sneakers", "collects antique jewelry", "collects luxury watches", "collects first edition books",
    "collects classic novels", "collects comic books", "enjoys fishing", "enjoys stargazing",
    "practices urban gardening", "enjoys rooftop gardening", "owns a microscope",
    "works on fridays", "drives a hybrid car", "has a pet dog", "is right-handed", "is a night owl", "uses contact lenses",
    "has a tattoo", "has a piercing", "travels domestically frequently", "is allergic to anything",
    "has lived in exactly three countries", "knows morse code", "makes homemade flans", "bakes bread at home",
    "is a tea enthusiast", "is a coffee connoisseur", "is a scotch connoisseur", "is a craft beer aficionado", "has a saltwater aquarium", "builds model airplanes", "owns a very old television",
    "enjoys logic puzzles", "uses a Windows laptop", "is a Linux enthusiast", "is a cybersecurity expert",
    "enjoys coding in Python", "streams on Twitch", "owns a smartwatch", "owns a 3D printer",
    "plays eSports competitively", "develops open-source software projects in their free time", "frequently participates in hackathons and coding competitions",
    "owns a high-end gaming PC with custom-built components", "regularly contributes to tech forums and online communities",
    "is an active member of a local robotics club", "creates augmented reality experiences for mobile applications",
    "works as a freelance web developer specializing in e-commerce sites", "hosts a popular podcast about emerging technologies",
    "maintains a personal blog focused on cybersecurity tips", "is a dedicated advocate for digital privacy and encryption",
    "creates large-scale murals for public art installations", "writes and illustrates their own graphic novels",
    "composes and performs experimental electronic music", "hosts regular workshops on creative writing",
    "is a member of a local theater group specializing in improv", "designs and sews custom cosplay costumes for conventions",
    "makes intricate hand-cut paper art for exhibitions", "hosts a YouTube channel dedicated to art tutorials",
    "creates bespoke furniture pieces from reclaimed wood", "is a professional photographer specializing in portrait photography",
    "trains for and competes in international triathlons", "is a certified yoga instructor teaching classes weekly",
    "plays as a goalkeeper for a local amateur soccer team", "participates in long-distance cycling events across the country",
    "is an avid mountain climber who has scaled several peaks", "mentors a youth basketball team on weekends",
    "competes in national level swimming championships", "practices and performs acrobatic dance routines",
    "enjoys deep-sea diving and exploring underwater caves",
    "collects rare and antique scientific instruments", "has a vast collection of first-edition science fiction novels",
    "owns an extensive assortment of vintage comic book memorabilia", "is passionate about collecting and restoring classic cars",
    "collects limited-edition art prints from contemporary artists", "has a curated collection of mid-century modern furniture",
    "collects historical artifacts related to ancient civilizations", "owns a significant collection of rare gemstones and minerals",
    "is an avid collector of autographed memorabilia from famous musicians", "has a specialized collection of handmade artisan pottery",
    "is an enthusiastic bird watcher who travels for rare sightings", "maintains a large, organic vegetable garden year-round",
    "volunteers for local environmental conservation projects", "regularly goes on multi-day backpacking trips in national parks",
    "enjoys kayaking and exploring remote waterways", "is a certified scuba diver with advanced underwater photography skills",
    "participates in citizen science projects related to wildlife monitoring", "is an expert in identifying and foraging wild edible plants",
    "enjoys camping and organizing outdoor survival workshops", "is dedicated to sustainable living and zero-waste practices",
    "is a culinary enthusiast who experiments with international cuisines", "hosts themed dinner parties featuring gourmet home-cooked meals",
    "is a wine connoisseur with a private cellar of vintage wines", "is a dedicated volunteer for local community service projects",
    "enjoys writing detailed reviews of new and classic board games", "is a chess master who participates in national tournaments",
    "hosts regular game nights featuring complex strategy games", "is an amateur astronomer who builds and uses custom telescopes",
    "writes in-depth travel guides for off-the-beaten-path destinations"
]



import os, io, subprocess, random, tempfile
from functools import lru_cache
from easydict import EasyDict as edict
import stat, shutil
from appdirs import user_data_dir
import pooch
import json

def get_vampire_path():
    path = shutil.which("vampire")
    if path:
        return path

    cache_dir = user_data_dir("vampire-wrapper")
    os.makedirs(cache_dir, exist_ok=True)

    vampire_path = pooch.retrieve(
        url="https://github.com/vprover/vampire/releases/download/v4.9casc2024/vampire",
        fname="vampire",
        path=cache_dir,
        known_hash=None  # Optional
    )

    os.chmod(vampire_path, os.stat(vampire_path).st_mode | stat.S_IEXEC)
    return vampire_path


VAMPIRE_PATH = get_vampire_path()

class Serializable(dict):
    """
    Drop-in mix-in: any subclass becomes automatically json.dumps-able.
    """

    def __setattr__(self, key, value):
        super().__setattr__(key, value)   # keep normal attribute behaviour
        self[key] = value                 # keep the dict in sync

    def __delattr__(self, key):
        super().__delattr__(key)
        self.pop(key, None)

    # nice-to-have helper
    def to_json(self, **dump_kw):
        return json.dumps(self, **dump_kw)

class ProofOutput(Serializable):
    def __init__(self,proof, input=''):
        self.proof = proof
        self.rules, self.indices = extract_inferences_and_formulas(proof)
        self.status = proof.split(' SZS status ',1)[-1].split(' for ')[0]
        self.sat = self.status == "Satisfiable"
        self.input = input

        if "Time limit" in self.status:
            self.status = "Time limit"
        if "Refutation not found" in self.status:
            self.status = "Refutation not found"

    def __str__(self):
        return f"{self.status}" + (f":{len(self.indices)}" if self.indices else "")
    __repr__=__str__

    def to_dict(self):
        return edict({k: getattr(self,k) for k in self.__dict__  if not k.startswith('__')})

    def to_json(self):
        return json.dumps(self.to_dict())


def split_clauses(x,prefix='axiom',name_prefix='',do_split=True):
    clauses=x.split('&\n')
    if any(a.count('(')!=a.count(')') for a in clauses):
        clauses=[x]
    return '\n'.join([f"fof({name_prefix}{i},{prefix},{c})." for i,c in enumerate(clauses)])+"\n"


def to_tptp(x,background='',problem='prem',neg='',mode='sat',use_hypothesis=True):
    mode={'sat':'axiom','proof':'conjecture'}[mode]
    premise = split_clauses(x.tptp,prefix=mode,name_prefix="p")
    if use_hypothesis:
        hypothesis = f"fof(hypothesis,{mode},{neg}({x.hyp_tptp}))."
    else:
        hypothesis=""
    return f"{background}\n{premise}\n{hypothesis}".replace('¿','?')

def run(expr, solver='vampire', proof=False, verbose=True):
    if not expr.strip().endswith(').') and not expr.strip().startswith('fof'):
        expr = f"fof(expr,axiom,{expr})."
    
    with tempfile.NamedTemporaryFile(mode="w", delete=False, dir="/dev/shm/") as f:
        f.write(expr)
        path = f.name
    
    try:
        cmd = [VAMPIRE_PATH, path, "--output_axiom_names", "on", "-t", "20s"]
        result = subprocess.run(cmd, text=True, capture_output=True)
        output = result.stdout if verbose else result.stdout.split(' SZS status ', 1)[-1].split(' for ')[0]
        return ProofOutput(output, input=expr)
    finally:
        os.remove(path)




def extract_inferences_and_formulas(proof):
    inferences,inputs=[],[]
    for x in proof.split('\n'):
        if not x.endswith(']'): 
            continue
        x=x.split('[')[-1].strip(']')        
        if x.startswith('input'):
            inputs.append(x.replace('input ',''))
        inferences.append(x.rsplit(' ',-1)[0])
    return inferences,inputs

def extract_inferences_and_formulas_tff(tff_statements):
    inference_types = list()
    formula_names = list()
    for line in tff_statements.strip().split("\n"):
        if "inference" in line:
            inference_types.append(line.split("inference(")[1].split(",")[0])
        if "file" in line:
            formula_names.append(line.split("file")[1].split(",")[1].split("'")[1].split('/')[-1])
    return inference_types, formula_names



import os, sys, re, random
import pandas as pd
from dataclasses import dataclass
import itertools
from functools import lru_cache, partial
from tqdm.auto import tqdm
from typing import Any
from anytree import Node, NodeMixin, LightNodeMixin, RenderTree
import psutil
from tqdm.auto import tqdm
import copy
from collections import defaultdict
import numpy as np
from easydict import EasyDict as edict
from functools import wraps
import time
from functools import lru_cache
from collections import deque
import random

class FlatList(list):
    def __iadd__(self, other):
        if isinstance(other, list):
            self.extend(other)
        else:
            self.append(other)
        return self

    def __add__(self, other):
        result = FlatList(self)
        if isinstance(other, list):
            result.extend(other)
        else:
            result.append(other)
        return result


def Constraint(constraint_str):
    def generated_function(x):
        conditions = constraint_str.split(',')
        for cond in conditions:
            i, j = map(int, cond.split('∉'))
            if x[i].render('eng') in x[j].render('eng'):
                return False
        return True
    return generated_function

def apply_to_all_args(f):
    def decorator(func):
        def wrapper(*args, **kwargs):
            new_args = [f(arg) for arg in args]
            return func(*new_args, **kwargs)
        return wrapper
    return decorator


def Substitution(template, lang=None):
    def replace_template(template, a):
        # Make numbers formattable 0 -> {0}
        wrap = lambda s: (re.sub(r'(\d+)', r'{\1}', s) if isinstance(s, str) else s)

        # Function to safely replace only unescaped '?'
        def replace_match(m):
            slot_idx = int(m.group(1))
            replacement = m.group(2)
            # The children args 'a' might be FastProduction nodes, so render them if needed
            arg_to_sub = a[slot_idx]
            if not isinstance(arg_to_sub, str):
                 arg_to_sub = arg_to_sub.render(lang)
            return re.sub(r'(?<!\\)\?', replacement, arg_to_sub)

        inner_replaced = re.sub(r"(\d+)\[\?←(.+?)\]", replace_match, template)
        
        # Render the remaining children before formatting
        rendered_args = [arg.render(lang) if not isinstance(arg, str) else arg for arg in a]
        output = wrap(inner_replaced).format(*rendered_args)

        # Convert escaped "\?" back to "?" after processing
        return output.replace(r'\?', '?')

    # The original Substitution returned a function that expected rendered strings.
    # The fix here is to make it robust enough to handle node objects too.
    def sub(*a, **ka):
        return replace_template(template, a)

    return sub

# Pre-compile regex patterns
NUMBER_PATTERN = re.compile(r'(\d+)')
SUBSTITUTION_PATTERN = re.compile(r"(\d+)\[\?←(.+?)\]")


default_preprocess_template = lambda s: (re.sub(r'(\d+)', r'{\1}', s) if type(s)==str and '←' not in s else s)

def init_grammar(langs, name='', preprocess_template=default_preprocess_template):
    class Rule:
        _instances = []
    
        @classmethod
        def init(cls, langs, name='', preprocess_template=preprocess_template):
            Rule.langs = langs
            Rule._instances = []
            Rule.preprocess_template = preprocess_template
            return langs
    
        def __init__(self, signature, *args, constraint=[], state_constraint=[], vars=dict(), weight=1):
            self.signature=signature
            self.name, self.args = self.parse_signature(signature)
            self.constraint = FlatList() + constraint
            self.state_constraint = FlatList() + state_constraint
            self.templates = {}
            self.weight=weight
            self.state = vars
            for lang, template in zip(Rule.langs, args*2):
                self.templates[lang] = Rule.preprocess_template(template)
            self.index = len(Rule._instances)
            Rule._instances.append(self)
    
        #def __getattr__(self,lang):
        #    return self.templates[lang]
    
        def parse_signature(self, s):
            name, args = re.match(r'(\w+)(?:\((.*?)\))?', s).groups()
            args = args.split(',') if args else []
            args = [a.strip() for a in args]
            return name, args
    
        @classmethod
        def get_rules(cls, name, sort=False, terminals=False,shuffle=False):
            instances = [x for x in cls._instances if x.name == name]
            if terminals:
                instances = [x for x in instances if not x.args]
            if shuffle:
                random.shuffle(instances)
            return instances
    
        @classmethod
        def start(cls):
            starts = cls.get_rules('start')
            if starts:
                return starts[0]
            else:
                return cls._instances[0]
    
        def __repr__(self):
            return f"RULE:{self.name}{self.args}"



        def __reduce__(self):
            def rebuild_rule(signature, constraint, state_constraint, state, weight, templates):
                # Logic to recreate the rule
                rule = Rule(signature, constraint=constraint, state_constraint=state_constraint, 
                            vars=state, weight=weight)
                rule.templates = templates
                return rule
                    
            return (rebuild_rule, (self.signature, self.constraint, self.state_constraint, 
                                  self.state, self.weight, self.templates))
                    
    R = Rule
    R.init(langs,name)

    classname = f'Rule_{id(object())}'
    cls_dict = dict(Rule.__dict__)
    cls_dict.pop('__dict__', None)
    cls_dict.pop('__weakref__', None)
    NewRule = type(classname, (object,), cls_dict)
    globals()[classname] = NewRule
    module = sys.modules[__name__]
    setattr(module, classname, NewRule)
    return NewRule

    


import copy
from collections import deque
from functools import lru_cache
import random
import re
from easydict import EasyDict as edict
import numpy as np
from unigram.grammar import Substitution

class FastProduction:
    """
    A unified, high-performance production node.
    """
    __slots__ = ('rule', 'type', 'parent', 'children', 'state', 'cache', 
                 '_depth', '_height')

    def __init__(self, rule=None, type=None, parent=None, children=None):
        self.rule = rule or ''
        self.type = type or (rule.name if rule else '')
        self.parent = parent
        
        self.cache = {}
 
        if children is not None:
            self.children = children
        elif self.rule:

            self.children = [
                FastProduction(type=arg, parent=self) 
                for arg in self.rule.args
            ]
        else:
            self.children = []

        self._depth = -1 
        self._height = -1
 

    def clone(self, memo=None):
        if memo is None: memo = {}
        if id(self) in memo: return memo[id(self)]
        new_obj = FastProduction(rule=self.rule, type=self.type)
        memo[id(self)] = new_obj
        new_obj.children = [child.clone(memo) for child in self.children]
        for child in new_obj.children:
            child.parent = new_obj
        return new_obj

    def check(self, mode=None):
        return True # Placeholder for your real constraint logic

    def update_height(self):
        if self._height != -1: return self._height
        self._height = 0 if not self.children else 1 + max((c.update_height() for c in self.children), default=0)
        return self._height
    @property
    def height(self):
        if self._height == -1: self.update_height()
        return self._height
    @property
    def depth(self):
        if self._depth != -1: return self._depth
        if self.parent is None: self._depth = 0
        else: self._depth = self.parent.depth + 1
        return self._depth
    @property
    def leaves(self):
        if not self.children: return [self]
        leaves_list, q = [], deque([self])
        while q:
            node = q.popleft()
            if not node.children: leaves_list.append(node)
            else: q.extendleft(reversed(node.children))
        return leaves_list
    @property
    def ancestors(self):
        anc, node = [], self.parent
        while node: anc.append(node); node = node.parent
        return tuple(anc)
    @property
    def siblings(self):
        return tuple(c for c in self.parent.children if c is not self) if self.parent else tuple()
    @property
    def descendants(self):
        if not self.children: return tuple()
        desc, q = [], deque(self.children)
        while q:
            node = q.popleft(); desc.append(node); q.extend(node.children)
        return tuple(desc)
    def render(self, lang=None):
        if lang in self.cache: return self.cache[lang]
        if not self.rule: return "#" + self.type
        template = self.rule.templates.get(lang)
        if template is None: return "#" + self.type
        if callable(template): args, template_func = self.children, template
        else:
            args = [c.render(lang) for c in self.children]
            if '?←' in template: template_func = Substitution(template, lang)
            else: template_func = template.format
        out = template_func(*args);
        if "#" not in out: self.cache[lang] = out
        return out
    __getitem__ = lambda self, key: self.children[key]
    __matmul__ = render
    __repr__ = lambda self: f"PROD:{self.type}" + (f"({', '.join(self.rule.args)})" if self.rule else "")

    def dict(self, use_cls=False):
        res = edict({l: self@l for l in self.rule.langs})
        if use_cls: res.cls = self
        return res



@lru_cache(maxsize=None)
def _precompute_height_bounds(Rule):
    """Computes min/max possible height for any non-terminal. Used by both modes."""
    non_terminals = list({r.name for r in Rule._instances})
    min_h, max_h = {nt: 999 for nt in non_terminals}, {nt: 0 for nt in non_terminals}
    for r in Rule._instances:
        if not r.args: min_h[r.name], max_h[r.name] = 0, 0
    for _ in range(len(non_terminals) + 1):
        for r in Rule._instances:
            if not r.args: continue
            child_mins = [min_h.get(arg, 999) for arg in r.args]
            child_maxs = [max_h.get(arg, 0) for arg in r.args]
            if all(h != 999 for h in child_mins):
                min_h[r.name] = min(min_h[r.name], 1 + max(child_mins))
            if all(h != 0 or max_h.get(arg, -1) !=-1 for arg, h in zip(r.args, child_maxs)): # handle cycles
                 max_h[r.name] = max(max_h[r.name], 1 + max(child_maxs))
    return min_h, max_h


def generate_recursive(start, depth=12, min_depth=None, n_iter=1000, bushiness_factor=0.5, **kwargs):
    """
    A fast, state-less generator that offers fine-grained control over tree topology.
    This version uses a true recursive helper function to avoid tree connection bugs.
    MODIFIED to perform weighted sampling of rules based on their .weight attribute.
    """
    Rule = type(start)
    min_heights, _ = _precompute_height_bounds(Rule)
    non_terminal_types = set(min_heights.keys())
    max_depth, min_depth = depth, min_depth or 0
    bushiness_factor = max(0.0, min(1.0, bushiness_factor))

    class _Node:
        __slots__ = ['rule', 'type', 'depth', 'children']
        def __init__(self, type, depth): self.type, self.depth, self.children = type, depth, []
        def __repr__(self): return f"_{self.type}(d={self.depth})"
        
    def _select_rule(rules):
        """Performs a weighted sample from a list of rules."""
        # This function assumes 'rules' is a non-empty list.
        # It defaults to a weight of 1.0 if the attribute is missing.
        weights = [getattr(r, 'weight', 1.0) for r in rules]
        
        # random.choices fails if all weights are 0. Fallback to uniform.
        if sum(weights) <= 0:
            return random.choice(rules)
            
        return random.choices(rules, weights=weights, k=1)[0]

    # --- The Recursive Tree-Building Helper Function ---
    def _build_tree(node, min_height_target):
        # Base case 1: Pruning - Impossible to meet target within max_depth
        if min_height_target > max_depth - node.depth:
            return False # Failure

        # Base case 2: Target met, fill randomly respecting only max_depth
        if min_height_target <= 0:
            potential_rules = [r for r in Rule.get_rules(node.type)
                               if node.depth + min_heights.get(r.name, 0) <= max_depth]
            if not potential_rules:
                return False # Failure
            
            node.rule = _select_rule(potential_rules)

            for arg in node.rule.args:
                child_node = _Node(type=arg, depth=node.depth + 1)
                node.children.append(child_node)
                if not _build_tree(child_node, 0): # Recurse with target 0
                    return False # Propagate failure
            return True # Success

        # Recursive Step: We have a depth target to meet-
        valid_rules = [r for r in Rule.get_rules(node.type)
                       if r.args and any(arg in non_terminal_types for arg in r.args)]
        if not valid_rules:
            return False # Failure
        
        node.rule = _select_rule(valid_rules)
        
        child_nodes = [_Node(type=arg, depth=node.depth + 1) for arg in node.rule.args]
        node.children = child_nodes
        
        potential_spine_indices = [i for i, child in enumerate(child_nodes) if child.type in non_terminal_types]
        if not potential_spine_indices: # All children are terminals, cannot continue spine
             for child_node in child_nodes:
                if not _build_tree(child_node, 0): return False
             return True

        spine_child_idx = random.choice(potential_spine_indices)
        
        new_spine_target = min_height_target - 1
        new_side_branch_target = int(np.floor(new_spine_target * bushiness_factor))
        
        success = True
        for i, child_node in enumerate(child_nodes):
            target = new_spine_target if i == spine_child_idx else new_side_branch_target
            if not _build_tree(child_node, target):
                success = False
                break
        return success

    # --- Main Generation Loop ---
    for _ in range(n_iter):
        root = _Node(type=start.name, depth=0)
        
        # Start the recursive generation process
        if _build_tree(root, min_depth):
            # If successful, convert the _Node tree to a FastProduction tree
            
            # Helper to collect all nodes from the valid _Node tree
            def collect_nodes(node):
                all_nodes = [node]
                for child in node.children:
                    all_nodes.extend(collect_nodes(child))
                return all_nodes

            # Now do the final conversion
            node_map = {}
            for node in reversed(collect_nodes(root)): # Post-order traversal
                prod = FastProduction(rule=node.rule, type=node.type, parent=None)
                children = [node_map[id(c)] for c in node.children]
                prod.children = children
                for child in children:
                    child.parent = prod
                node_map[id(node)] = prod
            
            return [node_map[id(root)]]

    return []


import random
from .. import Substitution, Constraint, generate, init_grammar

def tinypy_grammar(level=None):
    R = init_grammar(['py'])

    # -------------------------------------------------------------------------
    # 1. State Store & Reset Logic
    # -------------------------------------------------------------------------
    # This dictionary persists across generation calls, so we must reset it manually.
    state = {'assigned': {}, 'last': set(), 'loops': {}}
    chars = list("abcdefghijklmnopqrstuvwxyz")

    def reset_state(ctx_node):
        """Executed at the start of every program render to clean state."""
        state['assigned'].clear()
        state['last'].clear()
        state['loops'].clear()
        return "" # Render nothing, just side-effect

    # -------------------------------------------------------------------------
    # 2. Context-Sensitive Helpers
    # -------------------------------------------------------------------------
    def concat(*args):
        """Concatenates rendered children."""
        return "".join(a.render('py') for a in args)

    def render_init(ctx_node):
        """Creates a new variable assignment."""
        # Prefer unused variables
        candidates = [c for c in chars if c not in state['assigned']]
        v = random.choice(candidates if candidates else chars)
        d = str(random.randint(0, 255))
        state['assigned'][v] = d
        return f"{v} = {d}\n"

    def render_assign(v_node, e_node):
        """Updates 'last' variable tracking on assignment."""
        v, e = v_node.render('py'), e_node.render('py')
        state['last'].clear()
        state['last'].add(v)
        # Ensure it's marked as assigned
        state['assigned'][v] = state['assigned'].get(v, '0')
        return f"{v} = {e}\n"

    def get_var(ctx):
        """Returns a known variable, or falls back if none exist."""
        if state['assigned']: return random.choice(list(state['assigned'].keys()))
        return str(random.randint(0, 255))

    def get_last_var(ctx):
        """Returns the most recently modified variable (for print)."""
        if state['last']: return list(state['last'])[0]
        return get_var(ctx)

    def render_loop_math(ctx_node, mode):
        """Calculates Loop Bounds (Init/Step/Final) to ensure math coherence."""
        if mode == 'init':
            val = str(random.randint(0, 20))
            state['loops']['val'] = val
            return val
        
        init_val = int(state['loops'].get('val', '0'))
        # TinyPy specific step/count distribution
        step, count = random.choice([(1, 2), (2, 1), (2, 2), (2, 3), (3, 2)])
        state['loops']['step'] = str(step)
        
        if mode == 'final_less': return str(step * count + init_val - 1)
        if mode == 'final_greater': return str(init_val - step * count + 1)
        return "0"

    def render_while_var(ctx_node):
        """Selects a variable for a while loop and caches it."""
        v = get_var(ctx_node)
        # If we picked a digit by accident (fallback), pick a char
        if v.isdigit(): v = random.choice(chars)
        state['loops']['var'] = v
        state['loops']['val'] = state['assigned'].get(v, '0')
        return v

    def render_while_update(ctx_node, op):
        """Constructs update statement matching the loop variable."""
        v = state['loops'].get('var', 'i')
        s = state['loops'].get('step', '1')
        return f"{v} = {v} {op} {s}"

    # -------------------------------------------------------------------------
    # 3. Grammar Definition
    # -------------------------------------------------------------------------
    R('CTX', '') 
    R('RESET(CTX)', reset_state)

    # Terminals
    R('DIGIT', lambda: str(random.randint(0, 255)))
    R('VAR(CTX)', lambda x: random.choice(chars))
    for op in ['+', '-', '*', '/']: R('ARITH_OP', op)
    for op in ['<', '>', '<=', '>=', '!=', '==']: R('REL_OP', op)
    R('LOG_PREFIX', 'not ')
    for op in ['and', 'or']: R('LOG_INFIX', op)

    # Expressions
    R('EXPR_ID(CTX)', get_var)
    R('TERM(EXPR_ID)', '0'); R('TERM(DIGIT)', '0')
    R('EXPRESSION(TERM, ARITH_OP, TERM)', '0 1 2')
    R('ENCLOSED(EXPRESSION)', '(0)')

    R('DISP_ID(CTX)', get_last_var)
    R('DISP_EXPR(EXPR_ID, ARITH_OP, EXPR_ID)', '0 1 2')
    R('DISP_EXPR(EXPR_ID, ARITH_OP, DIGIT)', '0 1 2')

    # Initializations
    R('INIT(CTX)', render_init)
    # Unrolled max_init logic
    for n in [2, 3, 4, 5]:
        for k in range(1, n + 2):
            R(f"IDENT_INIT_{n}(" + ",".join(["INIT"]*k) + ")", concat)

    # Assignments
    R('SIMPLE_ARITH(ENCLOSED)', '0')
    R('SIMPLE_ARITH(SIMPLE_ARITH, ARITH_OP, ENCLOSED)', concat, weight=0.5)
    
    R('SIMPLE_ASSIGN(VAR, EXPRESSION)', render_assign)
    R('SIMPLE_ASSIGNS', '')
    R('SIMPLE_ASSIGNS(SIMPLE_ASSIGN)', '0')

    R('ADV_ASSIGN_TYPE(VAR, SIMPLE_ARITH)', render_assign)
    R('ADV_ASSIGN_TYPE(VAR, EXPRESSION)', render_assign)
    R('ADV_ASSIGNS', '')
    R('ADV_ASSIGNS(ADV_ASSIGN_TYPE)', '0')

    # Conditions
    R('COND_EXPR(EXPR_ID, REL_OP, EXPR_ID)', '0 1 2')
    R('COND_EXPR(EXPR_ID, REL_OP, DIGIT)', '0 1 2')
    R('COND(COND_EXPR)', '0')
    R('COND(LOG_PREFIX, COND_EXPR)', '01')
    R('ENCLOSED_COND(COND)', '(0)')
    R('CHAIN(ENCLOSED_COND)', '0')
    R('CHAIN(LOG_PREFIX, ENCLOSED_COND)', '01')
    R('CHAIN(CHAIN, LOG_INFIX, ENCLOSED_COND)', '0 1 2', weight=0.3)

    R('IF_BLK(COND)', 'if 0:\n')
    R('ELIF_BLK(COND)', 'elif 0:\n')
    R('ELSE_BLK', 'else:\n')
    R('ADV_IF(CHAIN)', 'if 0:\n')
    R('ADV_ELIF(CHAIN)', 'elif 0:\n')

    # Display
    R('DISPLAY(DISP_ID)', 'print(0)')
    R('ADV_DISP(DISPLAY)', '0')
    R('ADV_DISP(DISP_EXPR)', 'print(0)')

    # Loops
    R('FOR_INIT(CTX)', lambda x: render_loop_math(x, 'init'))
    R('FOR_FINAL(CTX)', lambda x: render_loop_math(x, 'final_less'))
    R('STEP(CTX)', lambda x: state['loops'].get('step', '1'))
    
    R('FOR_HEAD(EXPR_ID, FOR_INIT, FOR_FINAL, STEP)', 'for 0 in range(1, 2, 3):')
    R('FOR_HEAD(EXPR_ID, FOR_INIT, FOR_FINAL)', 'for 0 in range(1, 2):')
    R('FOR_LOOP(FOR_HEAD, DISPLAY)', '0\n\t1')
    R('ADV_FOR(FOR_HEAD, ADV_DISP)', '0\n\t1')
    R('ADV_FOR(FOR_LOOP)', '0')

    R('REL_LESS', '<'); R('REL_LESS', '<=')
    R('REL_GREATER', '>'); R('REL_GREATER', '>=')
    R('WHILE_VAR(CTX)', render_while_var)
    R('WH_FINAL_L(CTX)', lambda x: render_loop_math(x, 'final_less'))
    R('WH_FINAL_G(CTX)', lambda x: render_loop_math(x, 'final_greater'))
    R('WH_UPD_L(CTX)', lambda x: render_while_update(x, '+'))
    R('WH_UPD_G(CTX)', lambda x: render_while_update(x, '-'))

    R('WH_L(WHILE_VAR, REL_LESS, WH_FINAL_L, DISPLAY, WH_UPD_L)', 'while 0 1 2:\n\t3\n\t4')
    R('WH_G(WHILE_VAR, REL_GREATER, WH_FINAL_G, DISPLAY, WH_UPD_G)', 'while 0 1 2:\n\t3\n\t4')

    # Levels
    R('L1_1(IDENT_INIT_2, SIMPLE_ASSIGNS, ADV_DISP)', concat)
    R('L1_2(IDENT_INIT_3, ADV_ASSIGNS, ADV_DISP)', concat)

    R('L2_1(IDENT_INIT_2, IF_BLK, DISPLAY)', 
      lambda i, f, d: f"{i.render('py')}{f.render('py')}\t{d.render('py')}")
    R('L2_1(IDENT_INIT_2, IF_BLK, DISPLAY, ELIF_BLK, DISPLAY, ELSE_BLK, DISPLAY)', 
      lambda i, if_, d1, el_, d2, els, d3: f"{i.render('py')}{if_.render('py')}\t{d1.render('py')}\n{el_.render('py')}\t{d2.render('py')}\n{els.render('py')}\t{d3.render('py')}")
    R('L2_1(IDENT_INIT_2, IF_BLK, DISPLAY, ELSE_BLK, DISPLAY)', 
      lambda i, if_, d1, els, d2: f"{i.render('py')}{if_.render('py')}\t{d1.render('py')}\n{els.render('py')}\t{d2.render('py')}")

    R('L2_2(IDENT_INIT_5, ADV_ASSIGNS, ADV_IF, ADV_DISP)', 
      lambda i, a, f, d: f"{i.render('py')}{a.render('py')}{f.render('py')}\t{d.render('py')}")
    R('L2_2(IDENT_INIT_5, ADV_ASSIGNS, ADV_IF, ADV_DISP, ADV_ELIF, ADV_DISP, ELSE_BLK, ADV_DISP)', 
      lambda *x: f"{x[0].render('py')}{x[1].render('py')}{x[2].render('py')}\t{x[3].render('py')}\n{x[4].render('py')}\t{x[5].render('py')}\n{x[6].render('py')}\t{x[7].render('py')}")
    R('L2_2(IDENT_INIT_5, ADV_ASSIGNS, ADV_IF, ADV_DISP, ELSE_BLK, ADV_DISP)', 
      lambda *x: f"{x[0].render('py')}{x[1].render('py')}{x[2].render('py')}\t{x[3].render('py')}\n{x[4].render('py')}\t{x[5].render('py')}")

    R('L3_1(IDENT_INIT_2, FOR_LOOP)', concat)
    R('L3_2(IDENT_INIT_4, ADV_ASSIGNS, ADV_FOR)', concat)
    R('L4_1(IDENT_INIT_2, WH_L)', concat)
    R('L4_1(IDENT_INIT_2, WH_G)', concat)

    # Root - CRITICAL: Call r.render() to trigger RESET side-effects
    valid_levels = ["1.1", "1.2", "2.1", "2.2", "3.1", "3.2", "4.1"]
    selection = [level] if level in valid_levels else valid_levels

    for l in selection:
        rule_name = f"L{l.replace('.', '_')}"
        R(f'ALL(RESET, {rule_name})', lambda r, lvl: r.render('py') + lvl.render('py'))

    R('start(ALL)', '0')

    return R